@@ -49,7 +49,7 @@ GetOptions(\%args, qw(
     build=s host=s big-endian jit! enable-jit
     prefix=s bindir=s libdir=s mastdir=s
     relocatable make-install asan ubsan tsan
-    valgrind telemeh show-autovect git-cache-dir=s
+    valgrind telemeh dtrace show-autovect git-cache-dir=s
     show-autovect-failed:s),
 
     'no-optimize|nooptimize' => sub { $args{optimize} = 0 },
@@ -93,7 +93,7 @@ if ( $args{relocatable} && ($^O eq 'aix' || $^O eq 'openbsd') ) {
 }
 
 for (qw(coverage instrument static big-endian has-libtommath has-sha has-libuv
-        has-libatomic_ops asan ubsan tsan valgrind show-vec)) {
+        has-libatomic_ops asan ubsan tsan valgrind dtrace show-vec)) {
     $args{$_} = 0 unless defined $args{$_};
 }
 
@@ -426,6 +426,7 @@ push @cflags, '-fsanitize=thread' if $args{tsan};
 push @cflags, '-DWSL_BASH_ON_WIN' if wsl_bash_on_win();
 push @cflags, '-DDEBUG_HELPERS' if $args{debug};
 push @cflags, '-DMVM_VALGRIND_SUPPORT' if $args{valgrind};
+push @cflags, '-DMVM_DTRACE_SUPPORT' if $args{dtrace};
 push @cflags, '-DHAVE_TELEMEH' if $args{telemeh};
 push @cflags, '-DWORDS_BIGENDIAN' if $config{be}; # 3rdparty/sha1 needs it and it isnt set on mips;
 push @cflags, '-DMVM_HEAPSNAPSHOT_FORMAT=' . $config{heapsnapformat};
@@ -534,6 +535,7 @@ if ($config{cc} eq 'cl') {
 build::probe::C_type_bool(\%config, \%defaults);
 build::probe::computed_goto(\%config, \%defaults);
 build::probe::pthread_yield(\%config, \%defaults);
+build::probe::pthread_setname_np(\%config, \%defaults);
 build::probe::check_fn_malloc_trim(\%config, \%defaults);
 if ($^O eq 'aix') {
     build::probe::numbits(\%config, \%defaults);
@@ -1079,6 +1081,10 @@ A full list of options is displayed if you set C<TSAN_OPTIONS> to C<help=1>.
 
 Include Valgrind Client Requests for moarvm's own memory allocators.
 
+=item --dtrace
+
+Include DTrace trace points in various places.
+
 =item --ld <ld>
 
 Explicitly set the linker without affecting other configuration
