@@ -274,11 +274,11 @@ MVMObject * MVM_nativeref_lex_s(MVMThreadContext *tc, MVMuint16 outers, MVMuint1
 static MVMObject * lexref_by_name(MVMThreadContext *tc, MVMObject *type, MVMString *name, MVMint16 kind) {
     MVMFrame *cur_frame = tc->cur_frame;
     while (cur_frame != NULL) {
-        MVMLexicalRegistry *entry = MVM_get_lexical_by_name(tc, cur_frame->static_info, name);
-        if (entry) {
-            MVMint16 lex_kind = cur_frame->static_info->body.lexical_types[entry->value];
+        MVMuint32 idx = MVM_get_lexical_by_name(tc, cur_frame->static_info, name);
+        if (idx != MVM_INDEX_HASH_NOT_FOUND) {
+            MVMint16 lex_kind = cur_frame->static_info->body.lexical_types[idx];
             if (lex_kind == kind) {
-                return lex_ref(tc, type, cur_frame, entry->value, kind);
+                return lex_ref(tc, type, cur_frame, idx, kind);
             }
             /* If kind == LEXREF_ANY_INT we will allow any of the native int
              * types so we don't need functions for every single type. */
@@ -292,7 +292,7 @@ static MVMObject * lexref_by_name(MVMThreadContext *tc, MVMObject *type, MVMStri
                 case MVM_reg_uint16:
                 case MVM_reg_uint32:
                 case MVM_reg_uint64:
-                    return lex_ref(tc, type, cur_frame, entry->value, lex_kind);
+                    return lex_ref(tc, type, cur_frame, idx, lex_kind);
                 }
             }
             {
@@ -300,7 +300,7 @@ static MVMObject * lexref_by_name(MVMThreadContext *tc, MVMObject *type, MVMStri
                 char *waste[] = { c_name, NULL };
                 MVM_exception_throw_adhoc_free(tc, waste,
                     "Lexical with name '%s' has wrong type. real type %i wanted type %i",
-                        c_name, cur_frame->static_info->body.lexical_types[entry->value], kind);
+                        c_name, cur_frame->static_info->body.lexical_types[idx], kind);
             }
         }
         cur_frame = cur_frame->outer;
