@@ -79,7 +79,7 @@ typedef struct {
     MVMint32         handler_out_of_dynamic_scope;
 } LocatedHandler;
 
-static MVMint32 handler_can_handle(MVMFrame *f, MVMFrameHandler *fh, MVMint32 cat, MVMObject *payload) {
+static MVMint32 handler_can_handle(MVMFrame *f, MVMFrameHandler *fh, MVMuint32 cat, MVMObject *payload) {
     MVMuint32         category_mask = fh->category_mask;
     MVMuint64       block_has_label = category_mask & MVM_EX_CAT_LABELED;
     MVMuint64           block_label = block_has_label ? (uintptr_t)(f->work[fh->label_reg].o) : 0;
@@ -116,10 +116,10 @@ static MVMint32 search_frame_handlers_dyn(MVMThreadContext *tc, MVMFrame *f,
             }
         }
     } else {
-        MVMint32 num_handlers = f->spesh_cand
+        MVMuint32 num_handlers = f->spesh_cand
             ? f->spesh_cand->num_handlers
             : f->static_info->body.num_handlers;
-        MVMint32 pc;
+        MVMuint32 pc;
         if (f == tc->cur_frame)
             pc = (MVMuint32)(*tc->interp_cur_op - *tc->interp_bytecode_start);
         else
@@ -214,10 +214,10 @@ static MVMint32 search_frame_handlers_lex(MVMThreadContext *tc, MVMFrame *f,
         }
     }
     else {
-        MVMint32 num_handlers = f->spesh_cand
+        MVMuint32 num_handlers = f->spesh_cand
             ? f->spesh_cand->num_handlers
             : f->static_info->body.num_handlers;
-        MVMint32 pc;
+        MVMuint32 pc;
         if (f == tc->cur_frame)
             pc = (MVMuint32)(*tc->interp_cur_op - *tc->interp_bytecode_start);
         else
@@ -310,7 +310,7 @@ static LocatedHandler search_for_handler_from(MVMThreadContext *tc, MVMFrame *f,
                     lh.frame = f;
                     return lh;
                 }
-                f = f->caller;
+                f = MVM_LIKELY(f != tc->thread_entry_frame) ? f->caller : NULL;
             }
             return lh;
         case MVM_EX_THROW_LEXOTIC:
@@ -318,7 +318,7 @@ static LocatedHandler search_for_handler_from(MVMThreadContext *tc, MVMFrame *f,
                 lh = search_for_handler_from(tc, f, MVM_EX_THROW_LEX, cat, payload);
                 if (lh.frame != NULL)
                     return lh;
-                f = f->caller;
+                f = MVM_LIKELY(f != tc->thread_entry_frame) ? f->caller : NULL;
             }
             return lh;
         default:
@@ -344,6 +344,7 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
         else
             tc->last_payload = tc->instance->VMNull;
         /* Deliberate fallthrough to unwind below. */
+        MVM_FALLTHROUGH
 
     case MVM_EX_ACTION_GOTO:
         if (lh.jit_handler) {
@@ -395,6 +396,7 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
          * handler. */
         cur_frame->return_value = (MVMRegister *)&tc->last_handler_result;
         cur_frame->return_type = MVM_RETURN_OBJ;
+        cur_frame->return_address = *(tc->interp_cur_op);
         MVM_frame_special_return(tc, cur_frame, unwind_after_handler, cleanup_active_handler,
             ah, NULL);
 
@@ -547,7 +549,7 @@ MVMObject * MVM_exception_backtrace(MVMThreadContext *tc, MVMObject *ex_obj) {
         MVMuint32             offset = cur_op - MVM_frame_effective_bytecode(cur_frame);
         MVMBytecodeAnnotation *annot = MVM_bytecode_resolve_annotation(tc, &cur_frame->static_info->body,
                                             offset > 0 ? offset - 1 : 0);
-        MVMint32              fshi   = annot ? (MVMint32)annot->filename_string_heap_index : -1;
+        MVMuint32             fshi   = annot ? (MVMint32)annot->filename_string_heap_index : -1;
         char            *line_number = MVM_malloc(16);
         MVMString      *filename_str;
         snprintf(line_number, 16, "%d", annot ? annot->line_number : 1);
@@ -556,7 +558,7 @@ MVMObject * MVM_exception_backtrace(MVMThreadContext *tc, MVMObject *ex_obj) {
         annotations = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTHash);
 
         /* file */
-        filename_str = fshi >= 0 && fshi < cur_frame->static_info->body.cu->body.num_strings
+        filename_str = annot && fshi < cur_frame->static_info->body.cu->body.num_strings
              ? MVM_cu_string(tc, cur_frame->static_info->body.cu, fshi)
              : cur_frame->static_info->body.cu->body.filename;
         value = MVM_repr_box_str(tc, MVM_hll_current(tc)->str_box_type,
@@ -571,7 +573,8 @@ MVMObject * MVM_exception_backtrace(MVMThreadContext *tc, MVMObject *ex_obj) {
 
         /* row will contain "sub" and "annotations" */
         row = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTHash);
-        MVM_repr_bind_key_o(tc, row, k_sub, cur_frame->code_ref);
+        MVM_repr_bind_key_o(tc, row, k_sub,
+            cur_frame->code_ref ? cur_frame->code_ref : tc->instance->VMNull);
         MVM_repr_bind_key_o(tc, row, k_anno, annotations);
 
         MVM_repr_push_o(tc, arr, row);
@@ -601,20 +604,21 @@ MVMObject * MVM_exception_backtrace_strings(MVMThreadContext *tc, MVMObject *ex_
 
     MVMROOT(tc, ex, {
         arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    });
-    cur_frame = ex->body.origin;
 
-    MVMROOT2(tc, arr, cur_frame, {
-        MVMuint32 count = 0;
-        while (cur_frame != NULL) {
-            char *line = MVM_exception_backtrace_line(tc, cur_frame, count++,
-                ex->body.throw_address);
-            MVMString *line_str = MVM_string_utf8_decode(tc, tc->instance->VMString, line, strlen(line));
-            MVMObject *line_obj = MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr, line_str);
-            MVM_repr_push_o(tc, arr, line_obj);
-            cur_frame = cur_frame->caller;
-            MVM_free(line);
-        }
+        cur_frame = ex->body.origin;
+
+        MVMROOT2(tc, arr, cur_frame, {
+            MVMuint32 count = 0;
+            while (cur_frame != NULL) {
+                char *line = MVM_exception_backtrace_line(tc, cur_frame, count++,
+                    ex->body.throw_address);
+                MVMString *line_str = MVM_string_utf8_decode(tc, tc->instance->VMString, line, strlen(line));
+                MVMObject *line_obj = MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr, line_str);
+                MVM_repr_push_o(tc, arr, line_obj);
+                cur_frame = cur_frame->caller;
+                MVM_free(line);
+            }
+        });
     });
 
     return arr;
@@ -643,7 +647,13 @@ static void panic_unhandled_cat(MVMThreadContext *tc, MVMuint32 cat) {
             cat_name(tc, cat));
     }
     else {
-        fprintf(stderr, "No exception handler located for %s\n", cat_name(tc, cat));
+        if (tc->nested_interpreter)
+            fprintf(stderr, "An unhandled exception occurred in a native callback.\n"
+                    "This situation is not recoverable, and the program will terminate.\n"
+                    "The stack trace below helps indicate which library needs fixing.\n"
+                    "The exception was thrown at:\n");
+        else
+            fprintf(stderr, "No exception handler located for %s\n", cat_name(tc, cat));
         MVM_dump_backtrace(tc);
         if (crash_on_error)
             abort();
@@ -657,7 +667,9 @@ static void panic_unhandled_ex(MVMThreadContext *tc, MVMException *ex) {
     char *backtrace;
 
     /* If a debug session is running, notify the client. */
-    MVM_debugserver_notify_unhandled_exception(tc, ex);
+    MVMROOT(tc, ex, {
+        MVM_debugserver_notify_unhandled_exception(tc, ex);
+    });
 
     /* If it's a control exception, try promoting it to a catch one; use
      * the category name. */
