@@ -14,6 +14,10 @@ $Data::Dumper::Maxdepth = 1;
 # Download allkeys.txt from http://www.unicode.org/Public/UCA/latest/
 # and place into a folder named UCA under the UNIDATA directory.
 
+my $have_rakudo = `rakudo -e 'say "Hello world"'`;
+die "You need rakudo in your path to run this script\n"
+    unless $have_rakudo =~ /\AHello world/;
+
 my $DEBUG = $ENV{UCD2CDEBUG} // 0;
 
 binmode STDOUT, ':encoding(UTF-8)';
@@ -654,24 +658,21 @@ sub add_extent {
 }
 # Used in emit_codepoints_and_planes to push the codepoints name onto bitfield_index_lines
 sub ecap_push_name_line {
-    my ($name_lines, $name, $point, $bitfield_index_lines, $bytes, $index, $annotate_anyway) = @_;
+    my ($name_lines, $name, $point, $bitfield_index_lines, $index) = @_;
+    my $bytes;
     if (!defined $name) {
-        push @$bitfield_index_lines,
-            ($annotate_anyway
-                ? "/*$$index*/$point->{bitfield_index}/*$point->{code_str} */"
-                : "0"
-            );
+        push @$bitfield_index_lines, "0";
         push @$name_lines, "NULL";
     }
     else {
-        $$bytes += length($point->{name}) + 1; # length + 1 for the NULL
+        $bytes = length($point->{name}) + 1; # length + 1 for the NULL
         push @$bitfield_index_lines, "/*$$index*/$point->{bitfield_index}/* $point->{code_str} */";
         push @$name_lines, "/*$$index*/\"$point->{name}\"/* $point->{code_str} */";
     }
-    $$bytes += 2; # hopefully these are compacted since they are trivially aligned being two bytes
-    $$bytes += 8; # 8 for the pointer
-    $$index++;
-    return;
+    ++$$index;
+    $bytes += 2; # hopefully these are compacted since they are trivially aligned being two bytes
+    $bytes += 8; # 8 for the pointer
+    return $bytes;
 }
 sub emit_codepoints_and_planes {
     my ($first_point) = @_;
@@ -722,17 +723,10 @@ sub emit_codepoints_and_planes {
                 $toadd = $point;
                 $span_length = 0;
             }
-            my $usually = 1;  # occasionally change NULL to the name to cut name search time
             for (; 1 < $span_length; $span_length--) {
                 # catch up to last code
                 $last_point = $last_point->{next_point};
-                 # occasionally change NULL to the name to cut name search time
-                if ($last_point->{name} =~ / ^ [<] /x && $usually++ % 25) {
-                    ecap_push_name_line(\@name_lines, undef, $last_point, \@bitfield_index_lines, \$bytes, \$index, 1);
-                }
-                else {
-                    ecap_push_name_line(\@name_lines, $last_point->{name}, $last_point, \@bitfield_index_lines, \$bytes, \$index);
-                }
+                $bytes += ecap_push_name_line(\@name_lines, $last_point->{name}, $last_point, \@bitfield_index_lines, \$index);
             }
             $span_length = 0;
         }
@@ -747,7 +741,7 @@ sub emit_codepoints_and_planes {
             $toadd = $point;
         }
         for (; $last_code < $point->{code} - 1; $last_code++) {
-            ecap_push_name_line(\@name_lines, undef, $point, \@bitfield_index_lines, \$bytes, \$index);
+            $bytes += ecap_push_name_line(\@name_lines, undef, $point, \@bitfield_index_lines, \$index);
         }
 
         croak "$last_code  " . Dumper($point) unless $last_code == $point->{code} - 1;
@@ -759,7 +753,7 @@ sub emit_codepoints_and_planes {
         }
         $toadd = undef;
         # a normal codepoint that we don't want to compress
-        ecap_push_name_line(\@name_lines, $point->{name}, $point, \@bitfield_index_lines, \$bytes, \$index);
+        $bytes += ecap_push_name_line(\@name_lines, $point->{name}, $point, \@bitfield_index_lines, \$index);
         $last_code = $point->{code};
         $point->{main_index} = $index;
         $last_point = $point;
@@ -770,7 +764,7 @@ sub emit_codepoints_and_planes {
     # jnthn: Would it still use the same amount of memory to combine these tables? XXX
     $DB_SECTIONS->{BBB_codepoint_names} =
         "static const char *codepoint_names[$index] = {\n    ".
-            stack_lines(\@name_lines, ",", ",\n    ", 0, $WRAP_TO_COLUMNS).
+            stack_lines(\@name_lines, ",", ",\n    ", 0, 0).
             "\n};";
     $DB_SECTIONS->{BBB_codepoint_bitfield_indexes} =
         "static const MVMuint16 codepoint_bitfield_indexes[$index] = {\n    ".
@@ -911,7 +905,7 @@ sub emit_property_value_lookup {
 
 static MVMint32 MVM_unicode_get_property_int(MVMThreadContext *tc, MVMint64 codepoint, MVMint64 property_code) {
     MVMint32 result_val = 0; /* we'll never have negatives, but so */
-    MVMuint32 codepoint_row = MVM_codepoint_to_row_index(tc, codepoint);
+    MVMint32 codepoint_row = MVM_codepoint_to_row_index(tc, codepoint);
     MVMuint16 bitfield_row;
     /* If codepoint is not found in bitfield rows */
     if (codepoint_row == -1) {
@@ -935,7 +929,7 @@ static const char *bogus = "<BOGUS>"; /* only for table too short; return null s
 static const char* MVM_unicode_get_property_str(MVMThreadContext *tc, MVMint64 codepoint, MVMint64 property_code) {
     MVMuint32 switch_val = (MVMuint32)property_code;
     MVMint32 result_val = 0; /* we'll never have negatives, but so */
-    MVMuint32 codepoint_row;
+    MVMint32 codepoint_row;
     MVMuint16 bitfield_row = 0;
 
     if (switch_val == MVM_UNICODE_PROPERTY_BLOCK) {
@@ -1156,14 +1150,16 @@ sub emit_names_hash_builder {
     Okay not to be threadsafe since its value is deterministic
         and I don't care about the tiny potential for a memory leak
         in the event of a race condition. */
-static MVMUnicodeNameRegistry *codepoints_by_name = NULL;
+
+ /* static, so will be 0 initialised. */
+static MVMUniHashTable codepoints_by_name;
+
 static void generate_codepoints_by_name(MVMThreadContext *tc) {
     MVMint32 extent_index = 0;
     MVMint32 codepoint = 0;
     MVMint32 codepoint_table_index = 0;
     MVMint16 i = num_unicode_namealias_keypairs - 1;
 
-    MVMUnicodeNameRegistry *entry;
     for (; extent_index < MVM_NUM_UNICODE_EXTENTS; extent_index++) {
         MVMint32 length;
         codepoint = codepoint_extents[extent_index][0];
@@ -1178,11 +1174,11 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
                 for (; extent_span_index < length
                     && codepoint_table_index < MVM_CODEPOINT_NAMES_COUNT; extent_span_index++) {
                     const char *name = codepoint_names[codepoint_table_index];
-                    if (name) {
-                        MVMUnicodeNameRegistry *entry = MVM_malloc(sizeof(MVMUnicodeNameRegistry));
-                        entry->name = (char *)name;
-                        entry->codepoint = codepoint;
-                        HASH_ADD_KEYPTR(hash_handle, codepoints_by_name, name, strlen(name), entry);
+                    /* We want to skip various placeholder names that are duplicated:
+                     * <control> <CJK UNIFIED IDEOGRAPH> <CJK COMPATIBILITY IDEOGRAPH>
+                     * <surrogate> <TANGUT IDEOGRAPH> <private-use> */
+                    if (name && *name != '<') {
+                        MVM_uni_hash_insert(tc, &codepoints_by_name, name, codepoint);
                     }
                     codepoint++;
                     codepoint_table_index++;
@@ -1198,11 +1194,8 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
             /* Fate Span */
             case $FATE_SPAN: {
                 const char *name = codepoint_names[codepoint_table_index];
-                if (name) {
-                    MVMUnicodeNameRegistry *entry = MVM_malloc(sizeof(MVMUnicodeNameRegistry));
-                    entry->name = (char *)name;
-                    entry->codepoint = codepoint;
-                    HASH_ADD_KEYPTR(hash_handle, codepoints_by_name, name, strlen(name), entry);
+                if (name && *name != '<') {
+                    MVM_uni_hash_insert(tc, &codepoints_by_name, name, codepoint);
                 }
                 codepoint += length;
                 codepoint_table_index++;
@@ -1211,11 +1204,7 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
         }
     }
     for (; i >= 0; i--) {
-        entry = MVM_malloc(sizeof(MVMUnicodeNameRegistry));
-        entry->name = uni_namealias_pairs[i].name;
-        entry->codepoint =  uni_namealias_pairs[i].codepoint;
-        HASH_ADD_KEYPTR(hash_handle, codepoints_by_name,  uni_namealias_pairs[i].name,  uni_namealias_pairs[i].strlen, entry);
-
+        MVM_uni_hash_insert(tc, &codepoints_by_name, uni_namealias_pairs[i].name, uni_namealias_pairs[i].codepoint);
     }
 
 }
@@ -1591,7 +1580,7 @@ sub emit_unicode_property_value_keypairs {
             $done{"$propname$_"} ||= push @lines, $lines{$propname}->{$_};
         }
     }
-    my $out = "\nstatic MVMUnicodeNameRegistry **unicode_property_values_hashes;\n" .
+    my $out = "\nstatic MVMUniHashTable *unicode_property_values_hashes;\n" .
     "static const MVMUnicodeNamedValue unicode_property_value_keypairs[" . scalar(@lines) . "] = {\n" .
     "    " . stack_lines(\@lines, ",", ",\n    ", 0, $WRAP_TO_COLUMNS) . "\n" .
     "};";
@@ -1768,10 +1757,7 @@ sub read_file {
     my $fname = shift;
     open my $FILE, '<', $fname or croak "Couldn't open file '$fname': $!";
     binmode $FILE, ':encoding(UTF-8)';
-    my @lines = ();
-    while( <$FILE> ) {
-        push @lines, $_;
-    }
+    my @lines = <$FILE>;
     close $FILE;
     return \@lines;
 }
@@ -1780,6 +1766,9 @@ sub write_file {
     my ($fname, $contents) = @_;
     open my $FILE, '>', $fname or croak "Couldn't open file '$fname': $!";
     binmode $FILE, ':encoding(UTF-8)';
+    # Ensure generated files always end with a newline.
+    $contents .= "\n"
+        unless $contents =~ /\n\z/;
     print $FILE trim_trailing($contents);
     close $FILE;
     return;
@@ -2096,7 +2085,9 @@ sub Jamo {
         }
     }
     my $hs = join ',', @hangul_syllables;
-    my $out = `perl6 -e 'my \@cps = $hs; for \@cps -> \$cp { \$cp.chr.NFD.list.join(",").say };'`;
+    my $out = `rakudo -e 'my \@cps = $hs; for \@cps -> \$cp { \$cp.chr.NFD.list.join(",").say };'`;
+    die "Problem running rakudo to process Hangul syllables: \$? was $?"
+        if $?;
     my @out_lines = split "\n", $out;
     my $i = 0;
     for my $line (@out_lines) {
